"""
Test for fake_data_generator.py
"""
import numpy as np
from deepchem.utils.fake_data_generator import FakeGraphGenerator, generate_edge_index, remove_self_loops


def test_fake_graph_dataset():
    n_graphs = 10
    n_node_features = 5
    n_edge_features = 3
    n_classes = 2
    z_shape = 5

    # graph-level labels
    fgg = FakeGraphGenerator(min_nodes=3,
                             max_nodes=10,
                             n_node_features=n_node_features,
                             avg_degree=4,
                             n_edge_features=n_edge_features,
                             n_classes=n_classes,
                             task='graph',
                             z=z_shape)
    graphs = fgg.sample(n_graphs=n_graphs)

    assert len(graphs) == n_graphs
    assert np.unique(graphs.y).shape == (n_classes,)

    graph = graphs.X[0]
    assert graph.node_features.shape[1] == n_node_features
    assert graph.edge_features.shape[1] == n_edge_features
    assert graph.z.shape == (1, z_shape)

    # node-level labels
    fgg = FakeGraphGenerator(min_nodes=3,
                             max_nodes=10,
                             n_node_features=n_node_features,
                             avg_degree=4,
                             n_edge_features=n_edge_features,
                             n_classes=n_classes,
                             task='node',
                             z=z_shape)
    graphs = fgg.sample(n_graphs=n_graphs)

    assert len(graphs) == n_graphs

    graph = graphs.X[0]
    # graph.y contains node-labels and graph.node_features.shape[0]
    # holds number of nodes in that graph
    assert graph.y.shape[0] == graph.node_features.shape[0]
    assert graph.node_features.shape[1] == n_node_features
    assert graph.edge_features.shape[1] == n_edge_features
    assert graph.z.shape == (1, z_shape)


def test_generate_edge_index():
    n_nodes, avg_degree = 5, 3
    edge_indices = generate_edge_index(n_nodes, avg_degree, remove_loops=False)
    assert edge_indices.shape[0] == 2
    assert edge_indices.shape[1] == n_nodes * avg_degree


def test_remove_self_loops():
    edge_indices = np.array([[1, 2, 3], [1, 2, 4]])
    edge_indices = remove_self_loops(edge_indices)
    assert edge_indices.shape[0] == 2
    assert edge_indices.shape[1] == 1

    edge_indices = np.ones((2, 3))
    edge_indices = remove_self_loops(edge_indices)
    assert edge_indices.shape[0] == 2
    assert edge_indices.shape[1] == 0



"""
This file contains utilities for molecular docking.
"""
from typing import List, Optional, Tuple

import os
import numpy as np
from deepchem.utils.typing import RDKitMol
from deepchem.utils.pdbqt_utils import pdbqt_to_pdb


def write_vina_conf(protein_filename: str,
                    ligand_filename: str,
                    centroid: np.ndarray,
                    box_dims: np.ndarray,
                    conf_filename: str,
                    num_modes: int = 9,
                    exhaustiveness: Optional[int] = None) -> None:
    """Writes Vina configuration file to disk.

    Autodock Vina accepts a configuration file which provides options
    under which Vina is invoked. This utility function writes a vina
    configuration file which directs Autodock vina to perform docking
    under the provided options.

    Parameters
    ----------
    protein_filename: str
        Filename for protein
    ligand_filename: str
        Filename for the ligand
    centroid: np.ndarray
        A numpy array with shape `(3,)` holding centroid of system
    box_dims: np.ndarray
        A numpy array of shape `(3,)` holding the size of the box to dock
    conf_filename: str
        Filename to write Autodock Vina configuration to.
    num_modes: int, optional (default 9)
        The number of binding modes Autodock Vina should find
    exhaustiveness: int, optional
        The exhaustiveness of the search to be performed by Vina
    """
    with open(conf_filename, "w") as f:
        f.write("receptor = %s\n" % protein_filename)
        f.write("ligand = %s\n\n" % ligand_filename)

        f.write("center_x = %f\n" % centroid[0])
        f.write("center_y = %f\n" % centroid[1])
        f.write("center_z = %f\n\n" % centroid[2])

        f.write("size_x = %f\n" % box_dims[0])
        f.write("size_y = %f\n" % box_dims[1])
        f.write("size_z = %f\n\n" % box_dims[2])

        f.write("num_modes = %d\n\n" % num_modes)
        if exhaustiveness is not None:
            f.write("exhaustiveness = %d\n" % exhaustiveness)


def write_gnina_conf(protein_filename: str,
                     ligand_filename: str,
                     conf_filename: str,
                     num_modes: int = 9,
                     exhaustiveness: Optional[int] = None,
                     **kwargs) -> None:
    """Writes GNINA configuration file to disk.

    GNINA accepts a configuration file which provides options
    under which GNINA is invoked. This utility function writes a
    configuration file which directs GNINA to perform docking
    under the provided options.

    Parameters
    ----------
    protein_filename: str
        Filename for protein
    ligand_filename: str
        Filename for the ligand
    conf_filename: str
        Filename to write Autodock Vina configuration to.
    num_modes: int, optional (default 9)
        The number of binding modes GNINA should find
    exhaustiveness: int, optional
        The exhaustiveness of the search to be performed by GNINA
    kwargs:
        Args supported by GNINA documented here
        https://github.com/gnina/gnina#usage

    """

    with open(conf_filename, "w") as f:
        f.write("receptor = %s\n" % protein_filename)
        f.write("ligand = %s\n\n" % ligand_filename)

        f.write("autobox_ligand = %s\n\n" % protein_filename)

        if exhaustiveness is not None:
            f.write("exhaustiveness = %d\n" % exhaustiveness)
        f.write("num_modes = %d\n\n" % num_modes)

        for k, v in kwargs.items():
            f.write("%s = %s\n" % (str(k), str(v)))


def read_gnina_log(log_file: str) -> np.ndarray:
    """Read GNINA logfile and get docking scores.

    GNINA writes computed binding affinities to a logfile.

    Parameters
    ----------
    log_file: str
        Filename of logfile generated by GNINA.

    Returns
    -------
    scores: np.array, dimension (num_modes, 3)
        Array of binding affinity (kcal/mol), CNN pose score,
        and CNN affinity for each binding mode.

    """

    scores = []
    lines = open(log_file).readlines()
    mode_start = np.inf
    for idx, line in enumerate(lines):
        if line[:6] == '-----+':
            mode_start = idx
        if idx > mode_start:
            mode = line.split()
            score = [float(x) for x in mode[1:]]
            scores.append(score)

    return np.array(scores)


def load_docked_ligands(
        pdbqt_output: str) -> Tuple[List[RDKitMol], List[float]]:
    """This function loads ligands docked by autodock vina.

    Autodock vina writes outputs to disk in a PDBQT file format. This
    PDBQT file can contain multiple docked "poses". Recall that a pose
    is an energetically favorable 3D conformation of a molecule. This
    utility function reads and loads the structures for multiple poses
    from vina's output file.

    Parameters
    ----------
    pdbqt_output: str
        Should be the filename of a file generated by autodock vina's
        docking software.

    Returns
    -------
    Tuple[List[rdkit.Chem.rdchem.Mol], List[float]]
        Tuple of `molecules, scores`. `molecules` is a list of rdkit
        molecules with 3D information. `scores` is the associated vina
        score.

    Notes
    -----
    This function requires RDKit to be installed.
    """
    try:
        from rdkit import Chem
    except ModuleNotFoundError:
        raise ImportError("This function requires RDKit to be installed.")

    lines = open(pdbqt_output).readlines()
    molecule_pdbqts = []
    scores = []
    current_pdbqt: Optional[List[str]] = None
    for line in lines:
        if line[:5] == "MODEL":
            current_pdbqt = []
        elif line[:19] == "REMARK VINA RESULT:":
            words = line.split()
            # the line has format
            # REMARK VINA RESULT: score ...
            # There is only 1 such line per model so we can append it
            scores.append(float(words[3]))
        elif line[:6] == "ENDMDL":
            molecule_pdbqts.append(current_pdbqt)
            current_pdbqt = None
        else:
            # FIXME: Item "None" of "Optional[List[str]]" has no attribute "append"
            current_pdbqt.append(line)  # type: ignore

    molecules = []
    for pdbqt_data in molecule_pdbqts:
        pdb_block = pdbqt_to_pdb(pdbqt_data=pdbqt_data)
        mol = Chem.MolFromPDBBlock(str(pdb_block),
                                   sanitize=False,
                                   removeHs=False)
        molecules.append(mol)
    return molecules, scores


def prepare_inputs(protein: str,
                   ligand: str,
                   replace_nonstandard_residues: bool = True,
                   remove_heterogens: bool = True,
                   remove_water: bool = True,
                   add_hydrogens: bool = True,
                   pH: float = 7.0,
                   optimize_ligand: bool = True,
                   pdb_name: Optional[str] = None) -> Tuple[RDKitMol, RDKitMol]:
    """This prepares protein-ligand complexes for docking.

    Autodock Vina requires PDB files for proteins and ligands with
    sensible inputs. This function uses PDBFixer and RDKit to ensure
    that inputs are reasonable and ready for docking. Default values
    are given for convenience, but fixing PDB files is complicated and
    human judgement is required to produce protein structures suitable
    for docking. Always inspect the results carefully before trying to
    perform docking.

    Parameters
    ----------
    protein: str
        Filename for protein PDB file or a PDBID.
    ligand: str
        Either a filename for a ligand PDB file or a SMILES string.
    replace_nonstandard_residues: bool (default True)
        Replace nonstandard residues with standard residues.
    remove_heterogens: bool (default True)
        Removes residues that are not standard amino acids or nucleotides.
    remove_water: bool (default True)
        Remove water molecules.
    add_hydrogens: bool (default True)
        Add missing hydrogens at the protonation state given by `pH`.
    pH: float (default 7.0)
        Most common form of each residue at given `pH` value is used.
    optimize_ligand: bool (default True)
        If True, optimize ligand with RDKit. Required for SMILES inputs.
    pdb_name: Optional[str]
        If given, write sanitized protein and ligand to files called
        "pdb_name.pdb" and "ligand_pdb_name.pdb"

    Returns
    -------
    Tuple[RDKitMol, RDKitMol]
        Tuple of `protein_molecule, ligand_molecule` with 3D information.

    Note
    ----
    This function requires RDKit and OpenMM to be installed.
    Read more about PDBFixer here: https://github.com/openmm/pdbfixer.

    Examples
    --------
    >>> p, m = prepare_inputs('3cyx', 'CCC')

    >> p.GetNumAtoms()
    >> m.GetNumAtoms()

    >>> p, m = prepare_inputs('3cyx', 'CCC', remove_heterogens=False)

    >> p.GetNumAtoms()

    """

    try:
        from rdkit import Chem
        from rdkit.Chem import AllChem
        from pdbfixer import PDBFixer
        from simtk.openmm.app import PDBFile
    except ModuleNotFoundError:
        raise ImportError(
            "This function requires RDKit and OpenMM to be installed.")

    if protein.endswith('.pdb'):
        fixer = PDBFixer(protein)
    else:
        fixer = PDBFixer(url='https://files.rcsb.org/download/%s.pdb' %
                         (protein))

    if ligand.endswith('.pdb'):
        m = Chem.MolFromPDBFile(ligand)
    else:
        m = Chem.MolFromSmiles(ligand, sanitize=True)

    # Apply common fixes to PDB files
    if replace_nonstandard_residues:
        fixer.findMissingResidues()
        fixer.findNonstandardResidues()
        fixer.replaceNonstandardResidues()
    if remove_heterogens and not remove_water:
        fixer.removeHeterogens(True)
    if remove_heterogens and remove_water:
        fixer.removeHeterogens(False)
    if add_hydrogens:
        fixer.addMissingHydrogens(pH)

    PDBFile.writeFile(fixer.topology, fixer.positions, open('tmp.pdb', 'w'))
    p = Chem.MolFromPDBFile('tmp.pdb', sanitize=True)
    os.remove('tmp.pdb')

    # Optimize ligand
    if optimize_ligand:
        m = Chem.AddHs(m)  # need hydrogens for optimization
        AllChem.EmbedMolecule(m)
        AllChem.MMFFOptimizeMolecule(m)

    if pdb_name:
        Chem.rdmolfiles.MolToPDBFile(p, '%s.pdb' % (pdb_name))
        Chem.rdmolfiles.MolToPDBFile(m, 'ligand_%s.pdb' % (pdb_name))

    return (p, m)



"""
NCI dataset loader.
Original Author - Bharath Ramsundar
Author - Aneesh Pappu
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

NCI_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/nci_unique.csv"
NCI_TASKS = [
    'CCRF-CEM', 'HL-60(TB)', 'K-562', 'MOLT-4', 'RPMI-8226', 'SR', 'A549/ATCC',
    'EKVX', 'HOP-62', 'HOP-92', 'NCI-H226', 'NCI-H23', 'NCI-H322M', 'NCI-H460',
    'NCI-H522', 'COLO 205', 'HCC-2998', 'HCT-116', 'HCT-15', 'HT29', 'KM12',
    'SW-620', 'SF-268', 'SF-295', 'SF-539', 'SNB-19', 'SNB-75', 'U251',
    'LOX IMVI', 'MALME-3M', 'M14', 'MDA-MB-435', 'SK-MEL-2', 'SK-MEL-28',
    'SK-MEL-5', 'UACC-257', 'UACC-62', 'IGR-OV1', 'OVCAR-3', 'OVCAR-4',
    'OVCAR-5', 'OVCAR-8', 'NCI/ADR-RES', 'SK-OV-3', '786-0', 'A498', 'ACHN',
    'CAKI-1', 'RXF 393', 'SN12C', 'TK-10', 'UO-31', 'PC-3', 'DU-145', 'MCF7',
    'MDA-MB-231/ATCC', 'MDA-MB-468', 'HS 578T', 'BT-549', 'T-47D'
]


class _NCILoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "nci_unique.csv")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=NCI_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_nci(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'random',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load NCI dataset.

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    loader = _NCILoader(featurizer, splitter, transformers, NCI_TASKS, data_dir,
                        save_dir, **kwargs)
    return loader.load_dataset('nci', reload)



import tempfile
import unittest
import os

import numpy as np

from deepchem.utils import rdkit_utils


class TestRdkitUtil(unittest.TestCase):

    def setUp(self):
        # TODO test more formats for ligand
        current_dir = os.path.dirname(os.path.realpath(__file__))
        self.protein_file = os.path.join(
            current_dir, '../../feat/tests/data/3ws9_protein_fixer_rdkit.pdb')
        self.ligand_file = os.path.join(
            current_dir, '../../feat/tests/data/3ws9_ligand.sdf')

    def test_load_complex(self):
        complexes = rdkit_utils.load_complex(
            (self.protein_file, self.ligand_file),
            add_hydrogens=False,
            calc_charges=False)
        assert len(complexes) == 2

    def test_load_molecule(self):
        # adding hydrogens and charges is tested in dc.utils
        from rdkit.Chem.AllChem import Mol
        for add_hydrogens in (True, False):
            for calc_charges in (True, False):
                mol_xyz, mol_rdk = rdkit_utils.load_molecule(
                    self.ligand_file, add_hydrogens, calc_charges)
                num_atoms = mol_rdk.GetNumAtoms()
                self.assertIsInstance(mol_xyz, np.ndarray)
                self.assertIsInstance(mol_rdk, Mol)
                self.assertEqual(mol_xyz.shape, (num_atoms, 3))

    def test_get_xyz_from_mol(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")

        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        xyz2 = rdkit_utils.get_xyz_from_mol(mol)

        equal_array = np.all(xyz == xyz2)
        assert equal_array

    def test_add_hydrogens_to_mol(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")
        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        original_hydrogen_count = 0
        for atom_idx in range(mol.GetNumAtoms()):
            atom = mol.GetAtoms()[atom_idx]
            if atom.GetAtomicNum() == 1:
                original_hydrogen_count += 1

        assert mol is not None
        mol = rdkit_utils.add_hydrogens_to_mol(mol, is_protein=False)
        assert mol is not None
        after_hydrogen_count = 0
        for atom_idx in range(mol.GetNumAtoms()):
            atom = mol.GetAtoms()[atom_idx]
            if atom.GetAtomicNum() == 1:
                after_hydrogen_count += 1
        assert after_hydrogen_count >= original_hydrogen_count

    def test_apply_pdbfixer(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")
        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        original_hydrogen_count = 0
        for atom_idx in range(mol.GetNumAtoms()):
            atom = mol.GetAtoms()[atom_idx]
            if atom.GetAtomicNum() == 1:
                original_hydrogen_count += 1

        assert mol is not None
        mol = rdkit_utils.apply_pdbfixer(mol,
                                         hydrogenate=True,
                                         is_protein=False)
        assert mol is not None
        after_hydrogen_count = 0
        for atom_idx in range(mol.GetNumAtoms()):
            atom = mol.GetAtoms()[atom_idx]
            if atom.GetAtomicNum() == 1:
                after_hydrogen_count += 1
        assert after_hydrogen_count >= original_hydrogen_count

    def test_compute_charges(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")
        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=True)
        rdkit_utils.compute_charges(mol)

        has_a_charge = False
        for atom_idx in range(mol.GetNumAtoms()):
            atom = mol.GetAtoms()[atom_idx]
            value = atom.GetProp(str("_GasteigerCharge"))
            if value != 0:
                has_a_charge = True
        assert has_a_charge

    def test_load_molecule2(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")
        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        assert xyz is not None
        assert mol is not None

    def test_write_molecule(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")
        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)

        with tempfile.TemporaryDirectory() as tmp:
            outfile = os.path.join(tmp, "mol.sdf")
            rdkit_utils.write_molecule(mol, outfile)

            xyz, mol2 = rdkit_utils.load_molecule(outfile,
                                                  calc_charges=False,
                                                  add_hydrogens=False)

        assert mol.GetNumAtoms() == mol2.GetNumAtoms()
        for atom_idx in range(mol.GetNumAtoms()):
            atom1 = mol.GetAtoms()[atom_idx]
            atom2 = mol.GetAtoms()[atom_idx]
            assert atom1.GetAtomicNum() == atom2.GetAtomicNum()

    def test_merge_molecules_xyz(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")
        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        merged = rdkit_utils.merge_molecules_xyz([xyz, xyz])
        for i in range(len(xyz)):
            first_atom_equal = np.all(xyz[i] == merged[i])
            second_atom_equal = np.all(xyz[i] == merged[i + len(xyz)])
            assert first_atom_equal
            assert second_atom_equal

    def test_merge_molecules(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        ligand_file = os.path.join(current_dir,
                                   "../../dock/tests/1jld_ligand.sdf")
        xyz, mol = rdkit_utils.load_molecule(ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        num_mol_atoms = mol.GetNumAtoms()
        # self.ligand_file is for 3ws9_ligand.sdf
        oth_xyz, oth_mol = rdkit_utils.load_molecule(self.ligand_file,
                                                     calc_charges=False,
                                                     add_hydrogens=False)
        num_oth_mol_atoms = oth_mol.GetNumAtoms()
        merged = rdkit_utils.merge_molecules([mol, oth_mol])
        merged_num_atoms = merged.GetNumAtoms()
        assert merged_num_atoms == num_mol_atoms + num_oth_mol_atoms

    def test_merge_molecular_fragments(self):
        pass

    def test_strip_hydrogens(self):
        pass

    def test_all_shortest_pairs(self):
        from rdkit import Chem
        mol = Chem.MolFromSmiles("CN=C=O")
        valid_dict = {
            (0, 1): (0, 1),
            (0, 2): (0, 1, 2),
            (0, 3): (0, 1, 2, 3),
            (1, 2): (1, 2),
            (1, 3): (1, 2, 3),
            (2, 3): (2, 3)
        }
        assert rdkit_utils.compute_all_pairs_shortest_path(mol) == valid_dict

    def test_pairwise_ring_info(self):
        from rdkit import Chem
        mol = Chem.MolFromSmiles("c1ccccc1")
        predict_dict = rdkit_utils.compute_pairwise_ring_info(mol)
        assert all(pair == [(6, True)] for pair in predict_dict.values())
        mol = Chem.MolFromSmiles("c1c2ccccc2ccc1")
        predict_dict = rdkit_utils.compute_pairwise_ring_info(mol)
        assert all(pair == [(6, True)] for pair in predict_dict.values())
        mol = Chem.MolFromSmiles("CN=C=O")
        predict_dict = rdkit_utils.compute_pairwise_ring_info(mol)
        assert not predict_dict



"""Utility functions for working with PyTorch."""

import torch
from typing import Callable, Union


def get_activation(fn: Union[Callable, str]):
    """Get a PyTorch activation function, specified either directly or as a string.

    This function simplifies allowing users to specify activation functions by name.
    If a function is provided, it is simply returned unchanged.  If a string is provided,
    the corresponding function in torch.nn.functional is returned.
    """
    if isinstance(fn, str):
        return getattr(torch.nn.functional, fn)
    return fn



import time
import logging
import copy
from collections.abc import Sequence as SequenceCollection
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union, Sequence
import torch
import torch.nn as nn
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.optimizers import LearningRateSchedule
from deepchem.utils.typing import LossFn, OneOrMany

logger = logging.getLogger(__name__)


class ModularTorchModel(TorchModel):
    """ModularTorchModel is a subclass of TorchModel that allows for components to be
    pretrained and then combined into a final model. It is designed to be subclassed
    for specific models and is not intended to be used directly. There are 3 main differences
    between ModularTorchModel and TorchModel:

    - The build_components() method is used to define the components of the model.
    - The components are combined into a final model with the build_model() method.
    - The loss function is defined with the loss_func method. This may access the
      components to compute the loss using intermediate values from the network, rather
      than just the full forward pass output.

    Here is an example of how to use ModularTorchModel to pretrain a linear layer, load
    it into another network and then finetune that network:

    >>> import numpy as np
    >>> import deepchem as dc
    >>> import torch
    >>> n_samples = 6
    >>> n_feat = 3
    >>> n_hidden = 2
    >>> n_tasks = 6
    >>> pt_tasks = 3
    >>> X = np.random.rand(n_samples, n_feat)
    >>> y_pretrain = np.zeros((n_samples, pt_tasks)).astype(np.float32)
    >>> dataset_pt = dc.data.NumpyDataset(X, y_pretrain)
    >>> y_finetune = np.zeros((n_samples, n_tasks)).astype(np.float32)
    >>> dataset_ft = dc.data.NumpyDataset(X, y_finetune)
    >>> components = {'linear': torch.nn.Linear(n_feat, n_hidden),
    ... 'activation': torch.nn.ReLU(), 'head': torch.nn.Linear(n_hidden, n_tasks)}
    >>> model = torch.nn.Sequential(components['linear'], components['activation'],
    ... components['head'])
    >>> modular_model = dc.models.torch_models.modular.ModularTorchModel(model, components)
    >>> def example_loss_func(inputs, labels, weights):
    ...    return (torch.nn.functional.mse_loss(model(inputs), labels[0]) * weights[0]).mean()
    >>> modular_model.loss_func = example_loss_func
    >>> def example_model_build():
    ...     return torch.nn.Sequential(components['linear'], components['activation'],
    ... components['head'])
    >>> modular_model.build_model = example_model_build
    >>> pretrain_components = {'linear': torch.nn.Linear(n_feat, n_hidden),
    ... 'activation': torch.nn.ReLU(), 'head': torch.nn.Linear(n_hidden, pt_tasks)}
    >>> pretrain_model = torch.nn.Sequential(pretrain_components['linear'],
    ... pretrain_components['activation'], pretrain_components['head'])
    >>> pretrain_modular_model = dc.models.torch_models.modular.ModularTorchModel(pretrain_model,
    ... pretrain_components)
    >>> def example_pt_loss_func(inputs, labels, weights):
    ...     return (torch.nn.functional.mse_loss(pretrain_model(inputs), labels[0]) * weights[0]).mean()
    >>> pretrain_modular_model.loss_func = example_pt_loss_func
    >>> pt_loss = pretrain_modular_model.fit(dataset_pt, nb_epoch=1)
    >>> modular_model.load_pretrained_components(pretrain_modular_model, components=['linear'])
    >>> ft_loss = modular_model.fit(dataset_ft, nb_epoch=1)

    """

    def __init__(self, model: nn.Module, components: dict, **kwargs):
        """Create a ModularTorchModel.

        Parameters
        ----------
        model: nn.Module
            The model to be trained.
        components: dict
            A dictionary of the components of the model. The keys are the names of the
            components and the values are the components themselves.
        """

        self.model = model
        self.components = components
        super().__init__(self.model, self.loss_func, **kwargs)
        self.model.to(self.device)
        self.components = {
            k: v.to(self.device) for k, v in self.components.items()
        }

    def build_model(self) -> nn.Module:
        """Builds the final model from the components."""
        raise NotImplementedError("Subclass must define the components")

    def build_components(self) -> dict:
        """Creates the components dictionary, with the keys being the names of the
        components and the values being torch.nn.module objects."""
        raise NotImplementedError("Subclass must define the components")

    def loss_func(self, inputs: OneOrMany[torch.Tensor], labels: Sequence,
                  weights: Sequence) -> torch.Tensor:
        """Defines the loss function for the model which can access the components
        using self.components. The loss function should take the inputs, labels, and
        weights as arguments and return the loss."""
        raise NotImplementedError("Subclass must define the loss function")

    def freeze_components(self, components: List[str]):
        """Freezes or unfreezes the parameters of the specified components.

        Components string refers to keys in self.components.

        Parameters
        ----------
        components: List[str]
            The components to freeze.
        """
        for component in components:
            for param in self.components[component].parameters():
                param.requires_grad = False

    def unfreeze_components(self, components: List[str]):
        """Unfreezes the parameters of the specified components.

        Components string refers to keys in self.components.

        Parameters
        ----------
        components: List[str]
            The components to unfreeze.
        """
        for component in components:
            for param in self.components[component].parameters():
                param.requires_grad = True

    def load_pretrained_components(
            self,
            source_model: Optional['ModularTorchModel'] = None,
            checkpoint: Optional[str] = None,
            model_dir: Optional[str] = None,
            components: Optional[list] = None) -> None:
        """Modifies the TorchModel load_from_pretrained method to allow for loading
        from a ModularTorchModel and specifying which components to load.

        If the user does not a specify a source model, a checkpoint is used to load
        the weights. In this case, the user cannot specify which components to load
        because the components are not stored in the checkpoint. All layers will
        then be loaded if they have the same name and shape. This can cause issues
        if a pretrained model has similar but not identical layers to the model where
        a user may expect the weights to be loaded. ModularTorchModel subclasses
        should be written such that the components are atomic and will be preserved
        across as many tasks as possible. For example, an encoder may have varying
        input dimensions for different datasets, so the encoder should be written
        such that the input layer is not included in the encoder, allowing the
        encoder to be loaded with any input dimension.

        Parameters
        ----------
        source_model: Optional[ModularTorchModel]
            The model to load the weights from.
        checkpoint: Optional[str]
            The path to the checkpoint to load the weights from.
        model_dir: Optional[str]
            The path to the directory containing the checkpoint to load the weights.
        components: Optional[list]
            The components to load the weights from. If None, all components will be
            loaded.
        """

        # generate the source state dict
        if source_model is not None:
            source_state_dict = source_model.model.state_dict()
        elif checkpoint is not None:
            source_state_dict = torch.load(checkpoint)['model_state_dict']
        elif model_dir is not None:
            checkpoints = sorted(self.get_checkpoints(model_dir))
            source_state_dict = torch.load(checkpoints[0])['model_state_dict']
        else:
            raise ValueError(
                "Must provide a source model, checkpoint, or model_dir")

        if components is not None:  # load the specified components
            if source_model is not None:
                assignment_map = {
                    k: v
                    for k, v in source_model.components.items()
                    if k in components
                }
                assignment_map_copy = copy.deepcopy(
                    assignment_map)  # deep copy to avoid modifying source_model
                self.components.update(assignment_map_copy)
                self.model = self.build_model()
            else:
                raise ValueError(
                    "If loading from checkpoint, you cannot pass a list of components to load"
                )
        else:  # or all components with matching names and shapes
            model_dict = self.model.state_dict()
            assignment_map = {
                k: v
                for k, v in source_state_dict.items()
                if k in model_dict and v.shape == model_dict[k].shape
            }
            model_dict.update(assignment_map)
            self.model.load_state_dict(model_dict)

    def fit_generator(self,
                      generator: Iterable[Tuple[Any, Any, Any]],
                      max_checkpoints_to_keep: int = 5,
                      checkpoint_interval: int = 1000,
                      restore: bool = False,
                      variables: Optional[List[torch.nn.Parameter]] = None,
                      loss: Optional[LossFn] = None,
                      callbacks: Union[Callable, List[Callable]] = [],
                      all_losses: Optional[List[float]] = None) -> float:
        """Train this model on data from a generator. This method is similar to
        the TorchModel implementation, but it passes the inputs directly to the
        loss function, rather than passing them through the model first.  This
        enables the loss to be calculated from intermediate steps of the model
        and not just the final output.

        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in training steps.
            Set this to 0 to disable automatic checkpointing.
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        variables: list of torch.nn.Parameter
            the variables to train.  If None (the default), all trainable variables in
            the model are used.
        loss: function
            a function of the form f(outputs, labels, weights) that computes the loss
            for each batch.  If None (the default), the model's standard loss function
            is used.
        callbacks: function or list of functions
            one or more functions of the form f(model, step) that will be invoked after
            every step.  This can be used to perform validation, logging, etc.
        all_losses: Optional[List[float]], optional (default None)
            If specified, all logged losses are appended into this list. Note that
            you can call `fit()` repeatedly with the same list and losses will
            continue to be appended.

        Returns
        -------
        The average loss over the most recent checkpoint interval
        """

        if not isinstance(callbacks, SequenceCollection):
            callbacks = [callbacks]
        self._ensure_built()
        self.model.train()
        avg_loss = 0.0
        last_avg_loss = 0.0
        averaged_batches = 0
        if loss is None:
            loss = self._loss_fn
        if variables is None:
            optimizer = self._pytorch_optimizer
            lr_schedule = self._lr_schedule
        else:
            var_key = tuple(variables)
            if var_key in self._optimizer_for_vars:
                optimizer, lr_schedule = self._optimizer_for_vars[var_key]
            else:
                optimizer = self.optimizer._create_pytorch_optimizer(variables)
                if isinstance(self.optimizer.learning_rate,
                              LearningRateSchedule):
                    lr_schedule = self.optimizer.learning_rate._create_pytorch_schedule(
                        optimizer)
                else:
                    lr_schedule = None
                self._optimizer_for_vars[var_key] = (optimizer, lr_schedule)
        time1 = time.time()

        # Main training loop.

        for batch in generator:
            if restore:
                self.restore()
                restore = False
            inputs: OneOrMany[torch.Tensor]
            inputs, labels, weights = self._prepare_batch(batch)

            # Execute the loss function, accumulating the gradients.

            if isinstance(inputs, list) and len(inputs) == 1:
                inputs = inputs[0]

            optimizer.zero_grad()
            batch_loss = self.loss_func(inputs, labels, weights)
            batch_loss.backward()
            optimizer.step()
            if lr_schedule is not None:
                lr_schedule.step()
            self._global_step += 1
            current_step = self._global_step

            avg_loss += float(batch_loss)

            # Report progress and write checkpoints.
            averaged_batches += 1
            should_log = (current_step % self.log_frequency == 0)
            if should_log:
                avg_loss = float(avg_loss) / averaged_batches
                logger.info('Ending global_step %d: Average loss %g' %
                            (current_step, avg_loss))
                if all_losses is not None:
                    all_losses.append(avg_loss)
                # Capture the last avg_loss in case of return since we're resetting to 0 now
                last_avg_loss = avg_loss
                avg_loss = 0.0
                averaged_batches = 0

            if checkpoint_interval > 0 and current_step % checkpoint_interval == checkpoint_interval - 1:
                self.save_checkpoint(max_checkpoints_to_keep)
            for c in callbacks:
                c(self, current_step)
            if self.tensorboard and should_log:
                self._log_scalar_to_tensorboard('loss', batch_loss,
                                                current_step)
            if (self.wandb_logger is not None) and should_log:
                all_data = dict({'train/loss': batch_loss})
                self.wandb_logger.log_data(all_data, step=current_step)

        # Report final results.
        if averaged_batches > 0:
            avg_loss = float(avg_loss) / averaged_batches
            logger.info('Ending global_step %d: Average loss %g' %
                        (current_step, avg_loss))
            if all_losses is not None:
                all_losses.append(avg_loss)
            last_avg_loss = avg_loss

        if checkpoint_interval > 0:
            self.save_checkpoint(max_checkpoints_to_keep)

        time2 = time.time()
        logger.info("TIMING: model fitting took %0.3f s" % (time2 - time1))
        return last_avg_loss



"""
Data Structures used to represented molecules for convolutions.
"""
# flake8: noqa

import csv
import random
import numpy as np


def cumulative_sum_minus_last(l, offset=0):
    """Returns cumulative sums for set of counts, removing last entry.

    Returns the cumulative sums for a set of counts with the first returned value
    starting at 0. I.e [3,2,4] -> [0, 3, 5]. Note last sum element 9 is missing.
    Useful for reindexing

    Parameters
    ----------
    l: list
        List of integers. Typically small counts.
    """
    return np.delete(np.insert(np.cumsum(l, dtype=np.int32), 0, 0), -1) + offset


def cumulative_sum(l, offset=0):
    """Returns cumulative sums for set of counts.

    Returns the cumulative sums for a set of counts with the first returned value
    starting at 0. I.e [3,2,4] -> [0, 3, 5, 9]. Keeps final sum for searching.
    Useful for reindexing.

    Parameters
    ----------
    l: list
        List of integers. Typically small counts.
    """
    return np.insert(np.cumsum(l), 0, 0) + offset


class ConvMol(object):
    """Holds information about a molecules.

    Resorts order of atoms internally to be in order of increasing degree. Note
    that only heavy atoms (hydrogens excluded) are considered here.
    """

    def __init__(self, atom_features, adj_list, max_deg=10, min_deg=0):
        """
        Parameters
        ----------
        atom_features: np.ndarray
            Has shape (n_atoms, n_feat)
        adj_list: list
            List of length n_atoms, with neighor indices of each atom.
        max_deg: int, optional
            Maximum degree of any atom.
        min_deg: int, optional
            Minimum degree of any atom.
        """

        self.atom_features = atom_features
        self.n_atoms, self.n_feat = atom_features.shape
        self.deg_list = np.array([len(nbrs) for nbrs in adj_list],
                                 dtype=np.int32)
        self.canon_adj_list = adj_list
        self.deg_adj_lists = []
        self.deg_slice = []
        self.max_deg = max_deg
        self.min_deg = min_deg

        self.membership = self.get_num_atoms() * [0]

        self._deg_sort()

        # Get the degree id list (which corrects for min_deg)
        self.deg_id_list = np.array(self.deg_list) - min_deg

        # Get the size of each degree block
        deg_size = [
            self.get_num_atoms_with_deg(deg)
            for deg in range(self.min_deg, self.max_deg + 1)
        ]

        self.degree_list = []
        for i, deg in enumerate(range(self.min_deg, self.max_deg + 1)):
            self.degree_list.extend([deg] * deg_size[i])

        # Get the the start indices for items in each block
        self.deg_start = cumulative_sum(deg_size)

        # Get the node indices when they are reset when the degree changes
        deg_block_indices = [
            i - self.deg_start[self.deg_list[i]] for i in range(self.n_atoms)
        ]

        # Convert to numpy array
        self.deg_block_indices = np.array(deg_block_indices, dtype=np.int32)

    def get_atoms_with_deg(self, deg):
        """Retrieves atom_features with the specific degree"""
        start_ind = self.deg_slice[deg - self.min_deg, 0]
        size = self.deg_slice[deg - self.min_deg, 1]
        return self.atom_features[start_ind:(start_ind + size), :]

    def get_num_atoms_with_deg(self, deg):
        """Returns the number of atoms with the given degree"""
        return self.deg_slice[deg - self.min_deg, 1]

    def get_num_atoms(self):
        return self.n_atoms

    def _deg_sort(self):
        """Sorts atoms by degree and reorders internal data structures.
    
        Sort the order of the atom_features by degree, maintaining original order
        whenever two atom_features have the same degree.
        """
        old_ind = range(self.get_num_atoms())
        deg_list = self.deg_list
        new_ind = list(np.lexsort((old_ind, deg_list)))

        num_atoms = self.get_num_atoms()

        # Reorder old atom_features
        self.atom_features = self.atom_features[new_ind, :]

        # Reorder old deg lists
        self.deg_list = [self.deg_list[i] for i in new_ind]

        # Sort membership
        self.membership = [self.membership[i] for i in new_ind]

        # Create old to new dictionary. not exactly intuitive
        old_to_new = dict(zip(new_ind, old_ind))

        # Reorder adjacency lists
        self.canon_adj_list = [self.canon_adj_list[i] for i in new_ind]
        self.canon_adj_list = [[old_to_new[k]
                                for k in self.canon_adj_list[i]]
                               for i in range(len(new_ind))]

        # Get numpy version of degree list for indexing
        deg_array = np.array(self.deg_list)

        # Initialize adj_lists, which supports min_deg = 1 only
        self.deg_adj_lists = (self.max_deg + 1 - self.min_deg) * [0]

        # Parse as deg separated
        for deg in range(self.min_deg, self.max_deg + 1):
            # Get indices corresponding to the current degree
            rng = np.array(range(num_atoms))
            indices = rng[deg_array == deg]

            # Extract and save adjacency list for the current degree
            to_cat = [self.canon_adj_list[i] for i in indices]
            if len(to_cat) > 0:
                adj_list = np.vstack([self.canon_adj_list[i] for i in indices])
                self.deg_adj_lists[deg - self.min_deg] = adj_list.astype(
                    np.int32)

            else:
                self.deg_adj_lists[deg - self.min_deg] = np.zeros(
                    [0, deg], dtype=np.int32)

        # Construct the slice information
        deg_slice = np.zeros([self.max_deg + 1 - self.min_deg, 2],
                             dtype=np.int32)

        for deg in range(self.min_deg, self.max_deg + 1):
            if deg == 0:
                deg_size = np.sum(deg_array == deg)
            else:
                deg_size = self.deg_adj_lists[deg - self.min_deg].shape[0]

            deg_slice[deg - self.min_deg, 1] = deg_size
            # Get the cumulative indices after the first index
            if deg > self.min_deg:
                deg_slice[deg - self.min_deg,
                          0] = (deg_slice[deg - self.min_deg - 1, 0] +
                                deg_slice[deg - self.min_deg - 1, 1])

        # Set indices with zero sized slices to zero to avoid indexing errors
        deg_slice[:, 0] *= (deg_slice[:, 1] != 0)
        self.deg_slice = deg_slice

    def get_atom_features(self):
        """Returns canonicalized version of atom features.
    
        Features are sorted by atom degree, with original order maintained when
        degrees are same.
        """
        return self.atom_features

    def get_adjacency_list(self):
        """Returns a canonicalized adjacency list.
    
        Canonicalized means that the atoms are re-ordered by degree.
    
        Returns
        -------
        list
            Canonicalized form of adjacency list.
        """
        return self.canon_adj_list

    def get_deg_adjacency_lists(self):
        """Returns adjacency lists grouped by atom degree.
    
        Returns
        -------
        list
            Has length (max_deg+1-min_deg). The element at position deg is
            itself a list of the neighbor-lists for atoms with degree deg.
        """
        return self.deg_adj_lists

    def get_deg_slice(self):
        """Returns degree-slice tensor.
    
        The deg_slice tensor allows indexing into a flattened version of the
        molecule's atoms. Assume atoms are sorted in order of degree. Then
        deg_slice[deg][0] is the starting position for atoms of degree deg in
        flattened list, and deg_slice[deg][1] is the number of atoms with degree deg.
    
        Note deg_slice has shape (max_deg+1-min_deg, 2).
    
        Returns
        -------
        deg_slice: np.ndarray
            Shape (max_deg+1-min_deg, 2)
        """
        return self.deg_slice

    # TODO(rbharath): Can this be removed?
    @staticmethod
    def get_null_mol(n_feat, max_deg=10, min_deg=0):
        """Constructs a null molecules
    
        Get one molecule with one atom of each degree, with all the atoms
        connected to themselves, and containing n_feat features.
    
        Parameters
        ----------
        n_feat : int
            number of features for the nodes in the null molecule
        """
        # Use random insted of zeros to prevent weird issues with summing to zero
        atom_features = np.random.uniform(0, 1, [max_deg + 1 - min_deg, n_feat])
        canon_adj_list = [
            deg * [deg - min_deg] for deg in range(min_deg, max_deg + 1)
        ]

        return ConvMol(atom_features, canon_adj_list)

    @staticmethod
    def agglomerate_mols(mols, max_deg=10, min_deg=0):
        """Concatenates list of ConvMol's into one mol object that can be used to feed
            into tensorflow placeholders. The indexing of the molecules are preseved during the
            combination, but the indexing of the atoms are greatly changed.
    
        Parameters
        ----------
        mols: list
            ConvMol objects to be combined into one molecule.
        """

        num_mols = len(mols)

        # Combine the features, then sort them by (atom_degree, mol_index)
        atoms_by_deg = np.concatenate([x.atom_features for x in mols])
        degree_vector = np.concatenate([x.degree_list for x in mols], axis=0)
        # Mergesort is a "stable" sort, so the array maintains it's secondary sort of mol_index
        order = degree_vector.argsort(kind='mergesort')
        ordered = np.empty(order.shape, np.int32)
        ordered[order] = np.arange(order.shape[0], dtype=np.int32)
        all_atoms = atoms_by_deg[order]

        # Create a map from the original atom indices within each molecule to the
        # indices in the combined object.
        mol_atom_map = []
        index_start = 0
        for mol in mols:
            mol_atom_map.append(ordered[index_start:index_start +
                                        mol.get_num_atoms()])
            index_start += mol.get_num_atoms()

        # Sort all atoms by degree.
        # Get the size of each atom list separated by molecule id, then by degree
        mol_deg_sz = np.zeros([max_deg - min_deg + 1, num_mols], dtype=np.int32)
        for i, mol in enumerate(mols):
            mol_deg_sz[:, i] += mol.deg_slice[:, 1]

        # Get the final size of each degree block
        deg_sizes = np.sum(mol_deg_sz, axis=1)

        # Get the index at which each degree starts, not resetting after each degree
        # And not stopping at any specific molecule

        deg_start = cumulative_sum_minus_last(deg_sizes)

        # Get the tensorflow object required for slicing (deg x 2) matrix, with the
        # first column telling the start indices of each degree block and the
        # second colum telling the size of each degree block
        deg_slice = np.array(list(zip(deg_start, deg_sizes)))

        # Determine the membership (atom i belongs to molecule membership[i])
        membership = np.empty(all_atoms.shape[0], np.int32)
        for i in range(num_mols):
            membership[mol_atom_map[i]] = i

        # Initialize the new degree separated adjacency lists
        deg_adj_lists = [
            np.empty([deg_sizes[deg], deg], dtype=np.int32)
            for deg in range(min_deg, max_deg + 1)
        ]

        # Update the old adjacency lists with the new atom indices and then combine
        # all together
        for deg in range(min_deg, max_deg + 1):
            row = 0  # Initialize counter
            deg_id = deg - min_deg  # Get corresponding degree id

            # Iterate through all the molecules
            for mol_id in range(num_mols):
                # Get the adjacency lists for this molecule and current degree id
                nbr_list = mols[mol_id].deg_adj_lists[deg_id]

                # Correct all atom indices to the final indices, and then save the
                # results into the new adjacency lists
                if nbr_list.shape[0] > 0:
                    if nbr_list.dtype == np.int32:
                        final_id = mol_atom_map[mol_id][nbr_list]
                        deg_adj_lists[deg_id][row:(
                            row + nbr_list.shape[0])] = final_id
                        row += nbr_list.shape[0]
                    else:
                        for i in range(nbr_list.shape[0]):
                            for j in range(nbr_list.shape[1]):
                                deg_adj_lists[deg_id][
                                    row, j] = mol_atom_map[mol_id][nbr_list[i,
                                                                            j]]
                            # Increment once row is done
                            row += 1

        # Get the final aggregated molecule
        concat_mol = MultiConvMol(all_atoms, deg_adj_lists, deg_slice,
                                  membership, num_mols)
        return concat_mol


class MultiConvMol(object):
    """Holds information about multiple molecules, for use in feeding information
        into tensorflow. Generated using the agglomerate_mols function
    """

    def __init__(self, nodes, deg_adj_lists, deg_slice, membership, num_mols):
        self.nodes = nodes
        self.deg_adj_lists = deg_adj_lists
        self.deg_slice = deg_slice
        self.membership = membership
        self.num_mols = num_mols
        self.num_atoms = nodes.shape[0]

    def get_deg_adjacency_lists(self):
        return self.deg_adj_lists

    def get_atom_features(self):
        return self.nodes

    def get_num_atoms(self):
        return self.num_atoms

    def get_num_molecules(self):
        return self.num_mols


class WeaveMol(object):
    """Molecular featurization object for weave convolutions.

    These objects are produced by WeaveFeaturizer, and feed into
    WeaveModel. The underlying implementation is inspired by [1]_.


    References
    ----------
    .. [1] Kearnes, Steven, et al. "Molecular graph convolutions: moving beyond fingerprints." Journal of computer-aided molecular design 30.8 (2016): 595-608.
    """

    def __init__(self, nodes, pairs, pair_edges):
        self.nodes = nodes
        self.pairs = pairs
        self.num_atoms = self.nodes.shape[0]
        self.n_features = self.nodes.shape[1]
        self.pair_edges = pair_edges

    def get_pair_edges(self):
        return self.pair_edges

    def get_pair_features(self):
        return self.pairs

    def get_atom_features(self):
        return self.nodes

    def get_num_atoms(self):
        return self.num_atoms

    def get_num_features(self):
        return self.n_features
